@@ -35,10 +35,9 @@ use crate::loader::html::HtmlLoader;
 use crate::loader::json::JsonLoader;
 use crate::loader::Loader;
 use crate::worker::miner::WorkerMiner;
-use crate::worker::{IndexMsg, MainToWorkerMessage, Worker, WorkerToMainMessage};
+use crate::worker::{MainToWorkerMessage, Worker, WorkerToMainMessage};
 
 use crossbeam::channel::{self, Receiver, Sender};
-use crossbeam::deque::Injector as CrossbeamInjector;
 use glob::Pattern;
 use normalize_line_endings::normalized;
 use rustc_hash::FxHashMap;
@@ -85,10 +84,6 @@ fn get_default_num_pls_per_dir() -> u32 {
     1000
 }
 
-fn get_default_num_field_stores_per_dir() -> u32 {
-    1000
-}
-
 fn get_default_with_positions() -> bool {
     true
 }
@@ -116,9 +111,6 @@ pub struct MorselsIndexingConfig {
     #[serde(default = "get_default_num_pls_per_dir")]
     num_pls_per_dir: u32,
 
-    #[serde(default = "get_default_num_field_stores_per_dir")]
-    num_stores_per_dir: u32,
-
     #[serde(default = "get_default_with_positions")]
     with_positions: bool,
 }
@@ -133,7 +125,6 @@ impl Default for MorselsIndexingConfig {
             exclude: get_default_exclude_patterns(),
             loader_configs: get_default_loader_configs(),
             num_pls_per_dir: get_default_num_pls_per_dir(),
-            num_stores_per_dir: get_default_num_field_stores_per_dir(),
             with_positions: get_default_with_positions(),
         }
     }
@@ -182,7 +173,6 @@ struct MorselsIndexingOutputConfig {
     pl_names_to_cache: Vec<u32>,
     num_docs_per_block: u32,
     num_pls_per_dir: u32,
-    num_stores_per_dir: u32,
     with_positions: bool,
 }
 
@@ -218,9 +208,9 @@ pub struct Indexer {
     workers: Vec<Worker>,
     loaders: Vec<Box<dyn Loader>>,
     doc_infos: Arc<Mutex<DocInfos>>,
-    index_unit_queue: Arc<CrossbeamInjector<IndexMsg>>,
     tx_main: Sender<MainToWorkerMessage>,
     rx_main: Receiver<WorkerToMainMessage>,
+    rx_worker: Receiver<MainToWorkerMessage>,
     num_workers_writing_blocks: Arc<Mutex<usize>>,
     lang_config: MorselsLanguageConfig,
     is_dynamic: bool,
@@ -324,6 +314,7 @@ impl Indexer {
             Arc::new(FieldInfos::init(
                 field_infos_by_name,
                 config.fields_config.field_store_block_size,
+                config.fields_config.num_stores_per_dir,
                 output_folder_path,
             ))
         };
@@ -353,9 +344,7 @@ impl Indexer {
         ) = channel::bounded(config.indexing_config.num_threads);
         let (tx_main, rx_worker): (
             Sender<MainToWorkerMessage>, Receiver<MainToWorkerMessage>
-        ) = channel::bounded(config.indexing_config.num_threads);
-
-        let index_unit_queue = Arc::from(CrossbeamInjector::new());
+        ) = channel::bounded(32); // TODO may be a little arbitrary
 
         let expected_num_docs_per_thread =
             (config.indexing_config.num_docs_per_block / (config.indexing_config.num_threads as u32) * 2) as usize;
@@ -375,7 +364,6 @@ impl Indexer {
             let field_info_clone = Arc::clone(&field_infos);
             let indexing_config_clone = Arc::clone(&indexing_config);
             let num_workers_writing_blocks_clone = Arc::clone(&num_workers_writing_blocks);
-            let index_unit_queue = Arc::clone(&index_unit_queue);
 
             workers.push(Worker {
                 id: i as usize,
@@ -389,7 +377,6 @@ impl Indexer {
                         indexing_config_clone,
                         expected_num_docs_per_thread,
                         num_workers_writing_blocks_clone,
-                        index_unit_queue,
                     )
                 }),
             });
@@ -415,9 +402,9 @@ impl Indexer {
             workers,
             loaders,
             doc_infos,
-            index_unit_queue,
             tx_main,
             rx_main,
+            rx_worker,
             num_workers_writing_blocks,
             lang_config: config.lang_config,
             is_dynamic,
@@ -428,8 +415,6 @@ impl Indexer {
         };
         indexer.start_block_number = indexer.block_number();
 
-        indexer.make_workers_index(num_threads);
-
         indexer
     }
 
@@ -466,12 +451,6 @@ impl Indexer {
         ((self.doc_id_counter as f64) / (self.indexing_config.num_docs_per_block as f64)).floor() as u32
     }
 
-    fn make_workers_index(&self, n: usize) {
-        for _i in 0..n {
-            self.tx_main.send(MainToWorkerMessage::Index).expect("Failed to restart index phase");
-        }
-    }
-
     pub fn index_file(&mut self, input_folder_path_clone: &Path, path: &Path, relative_path: &Path) {
         let timestamp = if let Ok(metadata) = std::fs::metadata(path) {
             if let Ok(modified) = metadata.modified() {
@@ -503,12 +482,14 @@ impl Indexer {
         for loader in self.loaders.iter() {
             if let Some(loader_results) = loader.try_index_file(input_folder_path_clone, path, relative_path)
             {
-                for mut loader_result in loader_results {
-                    if self.index_unit_queue.len() > 100 { // TODO 100 may be a little arbitrary
-                        self.doc_miner.index_doc(self.doc_id_counter, loader_result.get_field_texts());
-                    } else {
-                        self.index_unit_queue.push(IndexMsg::Index { doc_id: self.doc_id_counter, loader_result });
-                    }
+                for loader_result in loader_results {
+                    self.tx_main.send(MainToWorkerMessage::Index {
+                        doc_id: self.doc_id_counter,
+                        loader_result,
+                    }).expect("Failed to send index msg to worker!");
+
+
+                    Self::try_index_doc(&mut self.doc_miner, &self.rx_worker, 30); // TODO 30 a little arbitrary?
 
                     self.dynamic_index_info.add_doc_to_external_id(external_id, self.doc_id_counter);
 
@@ -516,27 +497,15 @@ impl Indexer {
                     self.spimi_counter += 1;
 
                     if self.spimi_counter == self.indexing_config.num_docs_per_block {
-                        let mut num_workers_writing_blocks = self.num_workers_writing_blocks.lock().unwrap();
-                        // Minimally need 1 worker to take up this job still
-                        while *num_workers_writing_blocks == self.indexing_config.num_threads {
-                            drop(num_workers_writing_blocks);
-                            std::thread::sleep(std::time::Duration::from_millis(10));
-                            num_workers_writing_blocks = self.num_workers_writing_blocks.lock().unwrap();
-                        }
-
-                        let num_active_workers = self.indexing_config.num_threads - *num_workers_writing_blocks;
-                        for _i in 0..num_active_workers {
-                            self.index_unit_queue.push(IndexMsg::Stop);
-                        }
+                        Self::try_index_doc(&mut self.doc_miner, &self.rx_worker, 0);
 
                         let main_thread_block_index_results = self.doc_miner.get_results();
                         let block_number = self.block_number() - 1;
-                        self.write_block(
-                            main_thread_block_index_results, block_number, false, &mut * num_workers_writing_blocks
+
+                        self.merge_block(
+                            main_thread_block_index_results, block_number, false,
                         );
                         self.spimi_counter = 0;
-
-                        self.make_workers_index(num_active_workers - 1);
                     }
                 }
                 break;
@@ -544,6 +513,18 @@ impl Indexer {
         }
     }
 
+    fn try_index_doc(doc_miner: &mut WorkerMiner, rx_worker: &Receiver<MainToWorkerMessage>, until: usize) {
+        while rx_worker.len() > until {
+            if let Ok(msg) = rx_worker.try_recv() {
+                if let MainToWorkerMessage::Index { doc_id, mut loader_result } = msg {
+                    doc_miner.index_doc(doc_id, loader_result.get_field_texts());
+                } else {
+                    panic!("Unexpected message received @main thread");
+                }
+            }
+        }
+    }
+
     pub fn write_morsels_source_config(mut config: MorselsConfig, config_file_path: &Path) {
         config.indexing_config.loader_configs = config
             .indexing_config
@@ -562,59 +543,26 @@ impl Indexer {
             .unwrap();
     }
 
-    fn write_morsels_config(&mut self) {
-        let serialized = serde_json::to_string(&MorselsOutputConfig {
-            ver: MORSELS_VERSION,
-            last_doc_id: self.doc_id_counter,
-            indexing_config: MorselsIndexingOutputConfig {
-                loader_configs: std::mem::take(&mut self.loaders)
-                    .into_iter()
-                    .map(|loader| (loader.get_name(), loader))
-                    .collect(),
-                pl_names_to_cache: self.dynamic_index_info.pl_names_to_cache.clone(),
-                num_docs_per_block: self.indexing_config.num_docs_per_block,
-                num_pls_per_dir: self.indexing_config.num_pls_per_dir,
-                num_stores_per_dir: self.indexing_config.num_stores_per_dir,
-                with_positions: self.indexing_config.with_positions,
-            },
-            lang_config: &self.lang_config,
-            cache_all_field_stores: self.cache_all_field_stores,
-            field_infos: &self.field_infos,
-        })
-        .unwrap();
-
-        File::create(self.output_folder_path.join("morsels_config.json"))
-            .unwrap()
-            .write_all(serialized.as_bytes())
-            .unwrap();
-    }
-
-    fn is_deletion_only_run(&self) -> bool {
-        self.doc_id_counter == self.start_doc_id
-    }
-
     pub fn finish_writing_docs(mut self, instant: Option<Instant>) {
         #[cfg(debug_assertions)]
         println!("@finish_writing_docs");
 
-        for _i in 0..self.indexing_config.num_threads {
-            self.index_unit_queue.push(IndexMsg::Stop);
-        }
-
         let first_block = self.start_block_number;
         let mut last_block = self.block_number();
 
         if self.spimi_counter != 0 {
+            Self::try_index_doc(&mut self.doc_miner, &self.rx_worker, 0);
+
             #[cfg(debug_assertions)]
             println!("Writing extra last spimi block");
 
             let main_thread_block_index_results = self.doc_miner.get_results();
-            self.write_block(main_thread_block_index_results, last_block, true, &mut 0);
+            self.merge_block(main_thread_block_index_results, last_block, true);
             self.spimi_counter = 0;
         } else if !self.is_deletion_only_run() {
             last_block -= 1;
-            self.wait_on_all_workers();
         }
+        self.wait_on_all_workers();
 
         #[cfg(debug_assertions)]
         println!(
@@ -628,6 +576,28 @@ impl Indexer {
 
         // Merge spimi blocks
         // Go through all blocks at once
+        self.merge_blocks(first_block, last_block);
+
+        self.write_morsels_config();
+
+        self.dynamic_index_info.write(&self.output_folder_path, self.doc_id_counter);
+
+        if !self.is_deletion_only_run() {
+            spimireader::common::cleanup_blocks(first_block, last_block, &self.output_folder_path);
+        }
+
+        if let Some(now) = instant {
+            print_time_elapsed(now, "Blocks merged!");
+        }
+
+        self.terminate_all_workers();
+    }
+
+    fn is_deletion_only_run(&self) -> bool {
+        self.doc_id_counter == self.start_doc_id
+    }
+
+    fn merge_blocks(&mut self, first_block: u32, last_block: u32) {
         let num_blocks = last_block - first_block + 1;
         if self.is_dynamic {
             if self.delete_unencountered_external_ids {
@@ -659,20 +629,32 @@ impl Indexer {
                 &mut self.dynamic_index_info,
             );
         }
+    }
 
-        self.write_morsels_config();
+    fn write_morsels_config(&mut self) {
+        let serialized = serde_json::to_string(&MorselsOutputConfig {
+            ver: MORSELS_VERSION,
+            last_doc_id: self.doc_id_counter,
+            indexing_config: MorselsIndexingOutputConfig {
+                loader_configs: std::mem::take(&mut self.loaders)
+                    .into_iter()
+                    .map(|loader| (loader.get_name(), loader))
+                    .collect(),
+                pl_names_to_cache: self.dynamic_index_info.pl_names_to_cache.clone(),
+                num_docs_per_block: self.indexing_config.num_docs_per_block,
+                num_pls_per_dir: self.indexing_config.num_pls_per_dir,
+                with_positions: self.indexing_config.with_positions,
+            },
+            lang_config: &self.lang_config,
+            cache_all_field_stores: self.cache_all_field_stores,
+            field_infos: &self.field_infos,
+        })
+        .unwrap();
 
-        self.dynamic_index_info.write(&self.output_folder_path, self.doc_id_counter);
-
-        if !self.is_deletion_only_run() {
-            spimireader::common::cleanup_blocks(first_block, last_block, &self.output_folder_path);
-        }
-
-        if let Some(now) = instant {
-            print_time_elapsed(now, "Blocks merged!");
-        }
-
-        self.terminate_all_workers();
+        File::create(self.output_folder_path.join("morsels_config.json"))
+            .unwrap()
+            .write_all(serialized.as_bytes())
+            .unwrap();
     }
 }
 
