@@ -437,6 +437,9 @@ pub fn modify_blocks(
     dict_table_writer.flush().unwrap();
     dict_string_writer.flush().unwrap();
 
-    dynamic_index_info.last_pl_number = if new_pls_offset != 0 || new_pl == 0 { new_pl } else { new_pl - 1 };
-    dynamic_index_info.num_docs = new_num_docs as u32;
+    dynamic_index_info.last_pl_number = if new_pls_offset != 0 || new_pl == 0 {
+        new_pl
+    } else {
+        new_pl - 1
+    };
 }
