@@ -156,10 +156,12 @@ pub fn merge_blocks(
 
     pl_writer.flush(curr_pl_offset, indexing_config.pl_cache_threshold, &mut dynamic_index_info.pl_names_to_cache);
 
-    dynamic_index_info.last_pl_number =
-        if curr_pl_offset != 0 || curr_pl == 0 { curr_pl } else { curr_pl - 1 };
-    dynamic_index_info.num_docs = doc_id_counter;
-
     dict_table_writer.flush().unwrap();
     dict_string_writer.flush().unwrap();
+
+    dynamic_index_info.last_pl_number = if curr_pl_offset != 0 || curr_pl == 0 {
+        curr_pl
+    } else {
+        curr_pl - 1
+    };
 }
